# -*- coding: utf-8 -*-
"""
    This module provides the local data providers.
    It can be used from both templates and controllers to
    for data fetching.
"""
import os
import sys
import json
from os.path import abspath, splitext, isfile, join
import cherrypy
import yaml

this = sys.modules[__name__]

this.data_path = None
this.psycopg2_available = True

try:
    import psycopg2
    import psycopg2.extras
except ImportError:
    this.psycopg2_available = False


# this is a pointer to the module object instance itself.

def get_postgresql(provider_url):
    """ retrievede data from table matching the 'provider_url' """
    database_url = os.getenv('DATABASE_URL')
    if not database_url:    # There is no DB configured
        return None

    if not this.psycopg2_available:
        raise Exception('You must install the psycopg2 python package to use a database.')

    conn = psycopg2.connect(database_url)
    cur = conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
    try:
        cur.execute("SELECT * FROM %s;" % provider_url)
    except psycopg2.ProgrammingError as e:
        if e.pgcode == '42P01':  # UNDEFINED TABLE
            conn.close()
            return None
    resultset = cur.fetchall()
    data = []
    for row in resultset:
        data.append(row)
    cur.close()
    conn.close()
    return data


def get(provider_name, provider_key=None):
    """ Return an iterable object with data content.
    """
    parsed_url = urlparse(provider_name)
    if parsed_url.scheme != '':  # Schemes are note supported at this time
        return
    if parsed_url.path.startswith('/'):  # Absolute paths are not allowed
        return
    resource_path = abspath(join(this.data_path, parsed_url.path))
    filename, file_extension = splitext(resource_path)
    if file_extension == '':    # Attempt with all supported file type extentions
        search_path = []
        for extension in ['.yaml', '.json']:
            search_path.append(resource_path + extension)
    else:
        search_path = [resource_path]

    found_resource_path = None
    for path in search_path:
        if isfile(path):
            found_resource_path = path
            break

    # First attemtp to find a local provider, if not found, try the database
    if not found_resource_path:
        data = get_postgresql(provider_name)
        if data is None:
            raise cherrypy.HTTPError(
                404, "Unable to locate data provider for \'%s\'" % provider_name
                )

    filename, file_extension = splitext(found_resource_path)
    if file_extension == '.yaml':
        data = load_yaml(found_resource_path)
    if file_extension == '.json':
        data = json.load(found_resource_path)

    if provider_key:
        return data[provider_key]
    else:
        return data


def load_yaml(filename):
    multi_document_found = False
    with open(filename) as data_file:
        try:
            data = yaml.load(data_file)
        except yaml.composer.ComposerError:
            multi_document_found = True
    if multi_document_found:
        with open(filename) as data_file:
            data = [x for x in yaml.load_all(data_file) if x]
    return data


def set_base_dir(data_path):
    this.data_path = data_path


def data_title(provider_name):
    """ because this is a very frequently used item fetch """
    return get(provider_name, 'title')



#! /usr/bin/env python

"""QuickWeb

Usage:
    quickweb create <app_directory> <template_name> [--force]
    quickweb run [<app_directory>] [-l <listener_address>] [--no-logs]
    quickweb setup-cf-deployment [<app_directory>]
    quickweb setup-docker-deployment [<app_directory>]
    quickweb describe [<app_directory>]
    quickweb templates

Examples:

    quickweb create my-quickweb-app bootstrap-navbar-fixed-top

Options:
  -h --help     Show this screen.
"""
import os
import sys
from os.path import join, dirname
from docopt import docopt
from colorama import init
from . import app_manager, template_manager, version

init()

# Make sure we use the source directory for imports when running during development
script_dir = join(dirname(os.path.realpath(__file__)), '..')
sys.path.insert(0, script_dir)


def main():
    arguments = docopt(__doc__, version='QuickWeb %s' % version())
    print()

    if arguments['create']:
        app_manager.create(
            arguments['<app_directory>'], arguments['<template_name>'], arguments['--force'])

    if arguments['run']:
        app_manager.run(
            arguments['<app_directory>'], arguments['<listener_address>'], arguments['--no-logs']
        )

    if arguments['setup-cf-deployment']:
        app_manager.setup_cf_deployment(arguments['<app_directory>'])

    if arguments['setup-docker-deployment']:
        app_manager.setup_docker_deployment(arguments['<app_directory>'])

    if arguments['describe']:
        app_manager.describe(arguments['<app_directory>'])

    if arguments['templates']:
        template_manager.list()


if __name__ == "__main__":
    main()



#!/usr/bin/python
""" core engine package """

import os.path
import cherrypy


def version():
    """ Return quickweb engine version """

    quickweb_dir = os.path.dirname((os.path.realpath(__file__)))
    version_filename = os.path.join(quickweb_dir, 'version')
    with open(version_filename) as version_file:
        on_file_version = version_file.readline().strip('\r\n')

    return on_file_version

log = cherrypy.log  # noqa: E305



#!/usr/bin/python
""" Provides dynamic content using python controller """
from os.path import basename, join, splitext
from fnmatch import fnmatch
import imp
import warnings
import cherrypy

from quickweb import controller
from quickweb.events import on_event


class Feature(object):
    """ templates """

    def setup(self):
        on_event("content_file_found", lambda x, y: self.on_found_content_file(x, y))

    def on_found_content_file(self, content_root, content_name):
        """ A file was found """
        on_file_name = "*.py"

        # We only care if it matches our selection pattern
        if not fnmatch(basename(content_name), on_file_name):
            return

        url = '/' + content_name

        noext_name, ext = splitext(url)
        url = noext_name

        module_fname = join(content_root, content_name)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            controller_module = imp.load_source(module_fname, module_fname)
        controller.attach(url, controller_module.Controller())
        cherrypy.engine.autoreload.files.add(module_fname)



#!/usr/bin/python
""" Changes web content lookup directory """
import os
import sys
from os.path import isdir, join, exists, abspath
import cherrypy

from quickweb.colorhelper import print_error
from quickweb.events import invoke
from quickweb.startup import web_app_config


class Feature(object):
    """ content_directory """
    _content_directory = ""
    _static_directories = []

    def setup(self):
        """
        chdir() into the provided directory
        """
        content_directory = self._content_directory = "webroot"

        if not isdir(content_directory):
            print_error("Unable to find content directory '%s'" % content_directory)
            sys.exit(1)
        cherrypy.engine.autoreload.files.add(content_directory)

    def activate(self):
        """ Trigger the content dir/file_dir events """
        content_dir_len = len(self._content_directory)

        invoke("content_directory_set", self._content_directory)

        for root, dirs, files in os.walk(self._content_directory):
            for dir_name in dirs:
                relative_name = join(root, dir_name)[content_dir_len + 1:]
                static_file_flag = join(root, dir_name, '.static')
                if exists(static_file_flag):
                    self.setup_static_dir(relative_name)
                else:
                    invoke("content_dir_found", self._content_directory, relative_name)
            for file_name in files:
                relative_name = join(root, file_name)[content_dir_len + 1:]
                if not self.is_on_static_path(relative_name):
                    invoke("content_file_found", self._content_directory, relative_name)

    def setup_static_dir(self, content_name):
        """ A directory matched our dir specification """

        content_root = self._content_directory

        if self.is_on_static_path(content_name):
            return

        url = "/" + content_name
        # gerenate a static directory section for the cherrypy static dir tool
        static_conf = {
            url: {
                'tools.staticdir.dir': abspath(join(content_root, content_name)),
                'tools.staticdir.on': True,
                'tools.staticdir.content_types': {'svg': 'image/svg+xml'}
            }
        }
        web_app_config.update(static_conf)

        self._static_directories.append(content_name)

    def is_on_static_path(self, content_name):
        for static_dir in self._static_directories:
            if content_name.startswith(static_dir):
                return True
        return False



# !/usr/bin/python

from html.parser import HTMLParser
from os.path import basename, join, exists, dirname
from quickweb._tempfile import TemporaryDirectory
import os
import sys
import shutil
import io
import requests
import zipfile
from quickweb.colorhelper import info, print_error

TEMPLATES_REPO = 'https://github.com/OpenPipe/QuickWeb-templates/archive/master.zip'


class MyHTMLParser(HTMLParser):
    def __init__(self, url, template_directory):
        HTMLParser.__init__(self)
        self.template_directory = template_directory
        self.url = url
        self.replacement_map = {}

    def handle_starttag(self, tag, attrs):
        attrs = dict(attrs)
        if tag == 'link':
            self.fetch_static_asset(attrs['href'])
        if tag == 'script' and 'src' in attrs:
            self.fetch_static_asset(attrs['src'])

            # print("Encountered a start tag:", tag)

    def handle_endtag(self, tag):
        return
        print("Encountered an end tag :", tag)

    def handle_data(self, data):
        data = data.strip(' \r\n')
        if data:
            print("Encountered some data  :", data)

    def fetch_static_asset(self, url):
        print(self.template_directory, url)
        if url.split(':')[0] in ['http', 'https']:  # External resource
            print("WARNING: External resource", url, 'Skipping')
            return
        url = url.strip('\r\n ')
        fetch_url = self.url + url
        output_filename = basename(fetch_url)
        resource_extension = url.split('.')[-1]
        resource_map = {
            'js': join('static', 'js'),
            'css': join('static', 'css'),
            'ico': None,
        }
        resource_type = resource_map.get(resource_extension, None)
        if resource_type is None:
            print("WARNING: Ignoring resource", url)
            return
        output_dir = join(self.template_directory, resource_map[resource_extension])
        if output_dir is None:  # We don't fetch icons
            return
        print("Fetching %s ->  %s/%s" % (fetch_url, output_dir, output_filename))
        os.makedirs(output_dir, mode=0o755, exist_ok=True)
        resource = requests.get(fetch_url)

        with open(join(output_dir, output_filename), 'w') as static_file:
            static_file.write(resource.text)
        self.replacement_map[url] = '/'+join(resource_extension, output_filename)


class TemplateMaker:
    def __init__(self, url, template_directory):
        self.url = url
        if template_directory is None:
            template_directory = url.strip('/').split('/')[-1]
        self.template_directory = template_directory

    def make(self, force):
        if exists(self.template_directory):
            if force:
                shutil.rmtree(self.template_directory)
            else:
                sys.stderr.write(self.template_directory + ' already exists!\n\
                Please use --force if you want to overwrite.\n')
                sys.exit(2)

        # Fetch the page
        page = requests.get(self.url)
        parser = MyHTMLParser(self.url, self.template_directory)
        html = page.text

        # Fetch any static elements
        parser.feed(page.text)

        # Replace resource urls with our uniformed static locations
        for old, new in parser.replacement_map.items():
            print(old, new)
            html = html.replace(old, new)

        # Create the index.html
        webroot_dir = join(self.template_directory, 'webroot')
        if not exists(webroot_dir):
            os.makedirs(webroot_dir, mode=0o755)
        index_filename = join(webroot_dir, 'index.html')
        print("Creating ", index_filename)
        with open(index_filename, 'w') as index_file:
            index_file.write(html)

        print("\nMake completed.")
        print("You can now execute:\n\t quickweb run", self.template_directory)


def download_archive():
    print("** Downloading templates archive from", info(TEMPLATES_REPO))
    page = requests.get(TEMPLATES_REPO)
    file_like_object = io.BytesIO(page.content)
    templates_archive = zipfile.ZipFile(file_like_object)
    return templates_archive


def download_template(template_name, app_directory):
    templates_archive = download_archive()
    with TemporaryDirectory() as tmpdirname:
        templates_archive.extractall(tmpdirname)
        templates_archive_tmp = join(tmpdirname, 'QuickWeb-templates-master')
        template_dirs = os.listdir(templates_archive_tmp)
        if template_name not in template_dirs:
            print_error('Unable to find template %s !' % template_name)
            sys.exit(2)
        template_root = join(templates_archive_tmp, template_name)
        template_provides = [x for x in os.listdir(template_root)]
        print("** The template provides: %s" % info(str(template_provides)))
        shutil.copytree(template_root, app_directory)


def list():
    templates_archive = download_archive()
    print("** The following templates are available:")
    for x in templates_archive.infolist():
        if x.filename.count('/') == 2 and x.filename.endswith('/'):
            print("   "+info(basename(dirname(x.filename))))
    print("**")



# -*- coding: utf-8 -*-
"""
    This module provides the web engine core startup/application configuration logic
"""
import os
import sys
from os.path import abspath, join, dirname

import cherrypy
import quickweb
from quickweb import controller, data_provider
from quickweb.colorhelper import info, print_error, print_warn

web_app_config = {}


def setup_app(app_name, app_directory, no_logs):
    """ setup the application initial configuration """

    test_mode = os.getenv("TEST_MODE")
    if test_mode:
        print_warn('Running in TEST mode')

    app_directory = abspath(app_directory)

    os.chdir(app_directory)
    set_engine_config(test_mode, no_logs)
    setup_features()

    data_provider.set_base_dir(app_directory)
    cherrypy.tree.mount(controller.get_app_root(), config=web_app_config)


def setup_features():
    """ Call the features setup function """

    core_features = {
        'web': [
            'content_directory',
            'controllers',
            'templates'
            ]
        }

    imported_features = []
    for feature_type, feature_list in core_features.items():
        features_list_names = ', ' .join(feature_list)
        print("** Setting up {0} features {1}"
              .format(info(feature_type), info(features_list_names)))
        for feature_name in feature_list:
            script_dir = dirname(abspath(__file__))
            module_fname = join(script_dir, 'features', feature_type, feature_name+".py")

            feature_dict = {}
            with open(module_fname) as source_file:
                exec(compile(source_file.read(), module_fname, 'exec'), feature_dict)
            try:
                feature = feature_dict['Feature']()
            except KeyError:
                print_error("Feature module '%s' does not provide a Feature class!" % feature_name)
                sys.exit(1)
            try:
                feature.setup()
            except:  # NOQA: E722
                print_error("Failed setting up feature '%s' !" % feature_name)
                raise
            imported_features.append(feature)

        for feature in imported_features:
            if hasattr(feature, "activate"):
                feature.activate()


def set_engine_config(test_mode, no_logs):
    """ Set engine global config options """
    quickweb.is_production = os.getenv('QUICKWEB_PRODUCTION', False)

    # Enforce utf8 encoding
    app_config = {
        'tools.encode.on': True,
        'tools.encode.encoding': 'utf-8', 'tools.encode.errors': 'replace',
        'tools.trailing_slash.on': False,
        'tools.sessions.on': True,
        'checker.on': False
    }

    if test_mode:
        app_config.update(
            {
                'log.screen': False,
                'log.access_file': "access.log",
                'log.error_file': "error.log",
                'engine.autoreload.on': True
            }
        )
    # Disable logging when running with --no-logs
    if no_logs:
        app_config.update(
            {
                'log.screen': False,
                'log.access_file': None,
                'log.error_file': None,
            }
        )

    cherrypy.config.update(app_config)



# -*- coding: utf-8 -*-
from os.path import join, basename
from quickweb import controller, template
from glob import glob
from time import time
import imp
import cherrypy
import os
import logging.config
import warnings


# Set a default app root handler, to bind controllers until we find an
# "index" controller
class TemplateRenderer(object):  # Application Server Root place holder

    @controller.publish
    def index(self):
        path = controller.controller_path()[1]
        if path == '':
            html_template_filename = 'index.html'
        else:
            html_template_filename = path + ".html"
        return template.render(html_template_filename)


def setup_logging():
    LOG_CONF = {
        'version': 1,

        'formatters': {
            'void': {
                'format': ''
            },
            'standard': {
                'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s'
            },
        },
        'handlers': {
            'default': {
                'level': 'INFO',
                'class': 'logging.StreamHandler',
                'formatter': 'standard',
                'stream': 'ext://sys.stdout'
            },
            'cherrypy_console': {
                'level': 'INFO',
                'class': 'logging.StreamHandler',
                'formatter': 'void',
                'stream': 'ext://sys.stdout'
            },
            'cherrypy_access': {
                'level': 'INFO',
                'class': 'logging.handlers.RotatingFileHandler',
                'formatter': 'void',
                'filename': 'access.log',
                'maxBytes': 10485760,
                'backupCount': 20,
                'encoding': 'utf8'
            },
            'cherrypy_error': {
                'level': 'INFO',
                'class': 'logging.handlers.RotatingFileHandler',
                'formatter': 'void',
                'filename': 'errors.log',
                'maxBytes': 10485760,
                'backupCount': 20,
                'encoding': 'utf8'
            },
        },
        'loggers': {
            '': {
                'handlers': ['default'],
                'level': 'INFO'
            },
            'db': {
                'handlers': ['default'],
                'level': 'INFO',
                'propagate': False
            },
            'cherrypy.access': {
                'handlers': ['cherrypy_access'],
                'level': 'INFO',
                'propagate': False
            },
            'cherrypy.error': {
                'handlers': ['cherrypy_console', 'cherrypy_error'],
                'level': 'INFO',
                'propagate': False
            },
        }
    }
    logging.config.dictConfig(LOG_CONF)


def set_default_config():

    # Enforce utf8 encoding
    cherrypy.config.update(
        {
            'tools.encode.on': True,
            'tools.encode.encoding': 'utf-8', 'tools.encode.errors': 'replace'
        }
    )
    cherrypy.config.update({'tools.sessions.on': True})
    cherrypy.config.update({'environment': 'embedded'})
    setup_logging()


# Load and setup controllers
def load_controllers(app_name, controller_dir):
    print("*** Checking for controllers")
    for controller_file in glob(join(controller_dir, '*.py')):
        module_name = app_name + '.controllers.' + basename(controller_file)
        controller_path = basename(controller_file).split('.')[0]
        print("- Loading controller %s -> %s " %
              (module_name, controller_path))
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)
            handler = imp.load_source(module_name, controller_file)
        controller.attach(controller_path, handler.Controller())


def set_template_dirs(template_dirs, webroot_directory):
    for template_dirname in template_dirs:
        print("- Setting template directory", template_dirname)
    template.set_directories(webroot_directory, template_dirs)


def set_root_appnodes(webroot_directory):
    for html_filename in glob(join(webroot_directory, '*.html')):
        node_name = html_filename.split('/')[-1].split('.')[0]
        if node_name.startswith('_'):  # Dot not render template elements
            continue
        if node_name == 'index':  # Index is already handled by the root dir
            continue
        controller.attach('/' + node_name, TemplateRenderer())


# Setup static file serve handling for 'static' sub-dirs
def set_static_dirs(application, static_dirs):
    print("*** Checking for static dirs")
    for location in static_dirs:
        for static_dir in os.listdir(location):
            full_static_dir = join(location, static_dir)
            print("- Setting static dir %s -> %s" %
                  (static_dir, full_static_dir))
            conf = {}
            conf['/' + basename(static_dir)] = {'tools.staticdir.dir': full_static_dir,
                                                'tools.staticdir.on': True
                                                }
            application.merge(conf)


def setup(app_name, app_directory):

    quickweb_dir = os.path.dirname((os.path.realpath(__file__)))
    version_filename = os.path.join(quickweb_dir, '..', 'version')
    with open(version_filename) as version_file:
        version = version_file.readline().strip('\r\n')

    application_root = controller.set_approot(TemplateRenderer())
    app = cherrypy.Application(application_root, script_name=None, config=None)

    set_default_config()

    start_t = time()
    print('=' * 10 + " Starting app %s using QuickWeb %s " % (app_name, version) + '=' * 10)
    print('= Startup location', app_directory)
    static_dirs = []
    webroot_directory = join(app_directory, 'webroot')
    for location in ['template', 'webroot']:
        for root, dirs, files in os.walk(join(app_directory), location):
            for name in dirs:
                if name == 'static':
                    static_directory = join(join(root, name))
                    if static_directory not in static_dirs:
                        static_dirs.append(static_directory)

    set_static_dirs(app, static_dirs)
    set_root_appnodes(webroot_directory)
    load_controllers(app_name, join(app_directory, 'controllers'))
    stop_t = time()
    print('=' * 10 + ' Startup completed in %0.3f ms ' %
          ((stop_t - start_t) * 1000.0))

    return app



#!/usr/bin/python
""" EVent management functions """
import sys

this = sys.modules[__name__]
this.handlers_map = {}


def invoke(name, *args, **kwargs):
    handlers_list = this.handlers_map.get(name, [])
    #  print('invoked', name, len(handlers_list))
    for handler in handlers_list:
        handler(*args, **kwargs)


def on_event(name, handler):
    #  print('event added for', name)
    handlers_list = this.handlers_map.get(name, [])
    handlers_list.append(handler)
    this.handlers_map[name] = handlers_list



""" Provides some modules supportings functions """
from fnmatch import fnmatch


def fnmatch_list(filename, pattern_list):
    """ Check filename against a list of patterns using fnmatch """
    if type(pattern_list) != list:
        pattern_list = [pattern_list]
    for pattern in pattern_list:
        if fnmatch(filename, pattern):
            return True
    return False



# -*- coding: utf-8 -*-
import os
import re
import sys
import cherrypy

this = sys.modules[__name__]


class ApplicationNode:
    pass


this._app_root = ApplicationNode()

# Dictionary with URL to Name mapping to be displayed on navigation components
this._navigation = {}

# List of "place_after" constraints
this._navigation_place_after = []


def set_app_root(controller_obj):
    for key, value in this._app_root.__dict__.items():
        setattr(controller_obj, key, value)
    this._app_root = controller_obj


def get_app_root():
    return this._app_root


def find_path_base(controller_path_name):
    controller_path_name = controller_path_name.strip(os.sep)
    path_parts = controller_path_name.split(os.sep)

    path_root = this._app_root

    path_to_resource = path_parts[:-1]
    for part_name in path_to_resource:
        new_path_root = getattr(path_root, part_name, None)
        path_root = new_path_root
    return path_root


def attach(controller_path_name, controller_obj):
    """ Binds a controller object to an application URL """

    # Map OS path separator to URL separators (Needed on Windows)
    controller_path_name = controller_path_name.replace(os.sep, '/')

    controller_path_name = controller_path_name.strip('/')
    path_parts = controller_path_name.split("/")

    #  print('attach', controller_path_name, controller_obj, len(this.app_root.__dict__))
    path_root = this._app_root

    # If a path
    resource_name = path_parts[-1]
    path_to_resource = path_parts[:-1]
    for part_name in path_to_resource:
        new_path_root = getattr(path_root, part_name, None)
        if new_path_root is None:
            new_path_root = ApplicationNode()
            setattr(path_root, part_name, new_path_root)
        path_root = new_path_root

    # If we get a new root index, migrate current root resources to the new one
    if controller_path_name == '':
        set_app_root(controller_obj)
    else:
        setattr(path_root, resource_name, controller_obj)


def publish(func=None, *args):
    return cherrypy.expose(func, *args)


def redirect(url):
    raise cherrypy.HTTPRedirect(url)


def get_cookie(name, default_value=None):
    def unescape(s):
        m = re.match(r'^"(.*)"$', s)
        s = m.group(1) if m else s
        return s.replace("\\\\", "\\")

    if name in cherrypy.request.cookie:
        return unescape(cherrypy.request.cookie[name].value).decode('unicode-escape')
    else:
        return default_value


def get_session_value(name, default_value=None):
    if name in cherrypy.session:
        return cherrypy.session[name]
    else:
        return default_value


def set_session_value(name, value):
    cherrypy.session[name] = value


def set_cookie(name, value, path='/', max_age=3600, version=1):
    cookie = cherrypy.response.cookie
    cookie[name] = value.encode('unicode-escape')
    cookie[name]['path'] = path
    cookie[name]['max-age'] = max_age
    cookie[name]['version'] = version


def delete_cookie(name):
    cherrypy.response.cookie[name] = 'deleting'
    cherrypy.response.cookie[name]['expires'] = 0


def set_response(name, value):
    cherrypy.response.headers[name] = value


def method():
    return cherrypy.request.method


def params():
    return cherrypy.request.params


def controller_url():
    """
    @return: the current request url
    """
    return cherrypy.request.path_info


def controller_path():
    """
    @return: the current request path
    """
    return cherrypy.request.path_info.split('/')


def get_config():
    return cherrypy.config


def navigation_elements(start_path=None, using_placement=True):
    """
    Returns a list with { 'name': name, 'link': link}
    for each renderable ellement found on path
    """
    app_node = get_app_root()
    _navigation = app_node.__dict__.get('_navigation')
    if _navigation:
        return _navigation

    nav_list = []
    for leaf_name, leaf_obj in get_app_root().__dict__.iteritems():
        contains_index = hasattr(leaf_obj, 'index')
        if contains_index and leaf_name[0] != '_':
            nav_name = this._navigation.get(leaf_name, leaf_name)
            nav_list.append({'name': nav_name, 'link': leaf_name})

    if using_placement:
        for constraint_source, constraint_target in this._navigation_place_after:
            source_pos = [i for i, x in enumerate(nav_list) if x['link'] == constraint_source][0]
            target_pos = [i for i, x in enumerate(nav_list) if x['link'] == constraint_target][0]
            if source_pos < target_pos:
                swapping_element = nav_list.pop(source_pos)
                nav_list.insert(target_pos, swapping_element)

    return nav_list


def current_path():
    return controller_path()[1]


def helpers():
    """ Returns a dict with objects that should be available for templates """
    helpers_dict = {
        'navigation_elements': navigation_elements,
        'current_path': current_path
    }
    return helpers_dict


def set_navigation_info(*args, **kwargs):
    #  cherrypy.(str(args))
    #  cherrypy.log(str(kwargs))

    cherrypy.log('Setting navigation info %s %s' % (str(args), str(kwargs)))

    valid_kwargs = ['name', 'place_after']
    for k in kwargs:
        if k not in valid_kwargs:
            raise Exception('Unsupported parameter %s' % k)

    if 'name' in kwargs:
        this._navigation[current_path()] = kwargs['name']

    # While the code provides support for multiple types, for simplicity only 'place_after'
    # is implemented.
    constraint_target = kwargs.get('place_after')
    if constraint_target[0] != '/':
        constraint_target = '/' + constraint_target

    if not constraint_target:
        return

    was_found = [x for x in navigation_elements(using_placement=False)
                 if x['link'] == constraint_target]

    if not was_found:
        cherrypy.log("Unknown target "+constraint_target)
        return

    was_found = [x for x in this._navigation_place_after if x[2] == constraint_target]

    if was_found:
        raise Exception("Duplicated place_after for target '%s'" % constraint_target)

    this._navigation_place_after(
        (current_path(), constraint_target)
        )



import sys
import inspect
from termcolor import cprint, colored


def print_error(x):
    return cprint(x, 'red', attrs=['bold'], file=sys.stderr)


def print_info(x):
    return cprint(x, 'cyan')


def print_success(x):
    return cprint(x, 'green', attrs=['bold'])


def print_warn(x):
    return cprint(x, 'yellow', attrs=['bold'], file=sys.stderr)


def info(x):
    return colored(x, 'cyan')


def warning(x):
    return colored(x, 'yellow')


def success(x):
    return colored(x, 'green',  attrs=['bold'])


#  Based on http://stackoverflow.com/a/17412094
def print_template(template, **kwargs):
    "Usage: s(string, **locals())"
    if not kwargs:
        frame = inspect.currentframe()
        try:
            kwargs = frame.f_back.f_locals
        finally:
            del frame
        if not kwargs:
            kwargs = globals()

    print(template.format(**kwargs))



"""
"""
import os
import sys
from time import time
import cherrypy
import quickweb
from quickweb import startup
from quickweb.colorhelper import info


def run(app_directory, listener_address=None, no_logs=False, running_describe=False):
    """
    When an application is run, the following is performed:

    - Identify application root
        - Check for qwstart.py on
            - Startup script directory
            - Current working directory

    - Setup port number, if $PORT is not set, use a random port
    """
    start_t = time()    # Use for startup time calculation
    print("** Starting application %s using QuickWeb %s " % (
        info(app_directory), info(quickweb.version()))
         )
    startup_cwd = os.getcwd()

    # Check if beeing run from gunicorn
    is_gunicorn = "gunicorn" in os.environ.get("SERVER_SOFTWARE", "")
    if is_gunicorn:
        sys.stderr.write(
            "Quickweb provides it's own HTTP server module.\n"
            "Running from another HTTP server is not supported at this time\n"
            )
        sys.exit(1)

    # Identify the application root directory
    app_root_directory = app_directory or os.getcwd()

    startup.setup_app('app_name', app_root_directory, no_logs)
    if running_describe:
        return
    colored_elapsed_time = info("%0.2fms" % ((time() - start_t) * 1000.0))
    print('=' * 10 + ' Startup completed in ' + colored_elapsed_time)

    # Determine the HTTP listener port
    listener_port = int(os.getenv('PORT', 8080))
    if os.name == 'posix':
        socket_host = '0.0.0.0'
    else:
        socket_host = '127.0.0.1'
    if listener_address is not None:
        socket_host = listener_address
    cherrypy.config.update({'server.socket_host': socket_host})
    cherrypy.config.update({'server.socket_port': listener_port})

    # In some platforms signals are not available:
    if hasattr(cherrypy.engine, 'signals'):
        cherrypy.engine.signals.subscribe()
    cherrypy.engine.subscribe('stop', lambda: os.chdir(startup_cwd))

    cherrypy.engine.start()
    cherrypy.engine.block()



#!/usr/bin/python
""" Provides dynamic content using a template engine """

import yaml
import re

from os.path import basename, dirname, join, exists, getmtime, splitext
from jinja2 import TemplateNotFound, Environment, select_autoescape, meta
from jinja2 import BaseLoader, Markup, escape
from markdown import markdown

from quickweb import controller
from quickweb.events import on_event
from quickweb.modules import fnmatch_list


class Feature(object):
    """ templates """
    _template_engine = None

    def setup(self):
        """ config contains a list of config sections """
        on_event("content_directory_set", self.on_content_directory_set)
        on_event("content_file_found", lambda x, y: self.on_found_content_file(x, y))

    def on_content_directory_set(self, content_directory):
        self._template_engine = TemplateEngine(content_directory)

    def on_found_content_file(self, content_root, content_name):
        """ A file was found """
        on_file_name = ["*.html", "*.md"]
        file_exclude = "_*"

        # We only care if it matches our selection pattern
        if not fnmatch_list(basename(content_name), on_file_name):
            return

        if file_exclude and fnmatch_list(basename(content_name), file_exclude):
            return

        url = '/' + content_name
        noext_name, ext = splitext(url)
        url = noext_name

        # Map 'index' files to dirname/
        if basename(url) == 'index':
            url = dirname(url)

        template_controller = TemplateController(content_name, self._template_engine)
        controller.attach(url, template_controller)


class TemplateController(object):
    """ Render templates from a webroot dir """

    def __init__(self, path, template_engine):
        self._path = path
        self._template_engine = template_engine

    @controller.publish
    def index(self):
        return self._template_engine.render(self._path)


class TemplateEngine:
    """
    This classes provides a Jinja2 based template renderer, extended with the following:
        - Inject quickweb specific helper functions to be used on templates (e.g. current_path)
        - If there is a .yaml file matching the name, pass it's content as template variables
    """

    _engine = None
    _loader = None
    _template_base_path = None

    def __init__(self, template_base_path):
        self._template_base_path = template_base_path
        self._loader = QWTemplateLoader(template_base_path)
        self._engine = Environment(
            autoescape=select_autoescape(['html', 'xml']),
            loader=self._loader
            )
        self._engine.globals.update(controller.helpers())
        self._engine.filters['paragraphs'] = paragraphs
        self._engine.filters['markdown'] = lambda x: Markup(markdown(x))

    def render(self, name):

        template = self._engine.get_template(name)
        # Load .yaml data for all referenced templates
        referenced_templates = self._loader.referenced_templates(name)
        # Load .yaml data for the requested template
        referenced_templates.append(name)
        yaml_variables = {}
        for template_name in referenced_templates:
            noext_name, ext = splitext(template_name)
            yaml_data_filename = noext_name + '.yaml'
            try:
                yaml_renderer = self._engine.get_template(yaml_data_filename)
            except TemplateNotFound:
                pass
            else:
                yaml_data = yaml.load(yaml_renderer.render())
                yaml_variables.update(dict(yaml_data))

        return template.render(**yaml_variables)


class QWTemplateLoader(BaseLoader):
    """
    The QWTemplateLoader loads plain Jinja2 templates from files, and augments them
    with the QuickWeb application specific behavior:
        - If the template does not start with a '_' and there is a '_base.html' file, change the
        memory loaded template source to be extended from '_base.html'
        - Identify other templates included/extended from the loded one, this list will be required
        during rendering, because the .yaml content bind to every included template must be
        provided.
    """
    _included_templates = {}    # List of templates included by a specific template

    def __init__(self, path):
        self.path = path

    def get_source(self, environment, template):
        template_fn, template_ext = splitext(template)
        path = join(self.path, template)
        if not exists(path):
            raise TemplateNotFound(template)
        mtime = getmtime(path)
        with open(path, 'rb') as f:
            source = f.read().decode('utf-8')
        if template_ext == '.md':
            source = markdown(source, extensions=['markdown.extensions.attr_list'])
        # If the .html does not extend other template, check if there is _base.html .
        # We will extend the template file from _base.html if present.
        if template_ext in ['.html', '.md'] and not re.match(r'^\s*{% extends "', source) \
                and not basename(template)[0] == '_':
            _base_filename = join(self.path, '_base.html')
            if exists(_base_filename):
                extended_source = '{% extends "_base.html" %}\n'
                extended_source += '{% block content %}'
                extended_source += source
                extended_source += '{% endblock %}'
                source = extended_source

        # Set dependencies for this template
        referenced_templates = self._find_referenced_templates(template, source, environment)
        self._included_templates[template] = referenced_templates

        return source, path, lambda: mtime == getmtime(path)

    def _find_referenced_templates(self, name, source, environment):
        try:
            ast = environment.parse(source)
        except:  # NOQA: E722
            print("Error rendering template "+name)
            raise
        references = [x for x in meta.find_referenced_templates(ast)]
        for name in references:
            with open(join(self.path, name), 'rb') as f:
                source = f.read().decode('utf-8')
                references.extend(self._find_referenced_templates(name, source, environment))
        return references

    def referenced_templates(self, name):
        return self._included_templates.get(name)


def paragraphs(input):
    """ Split lines and return them enclosed in HTML paragraphs """
    html_p_list = list(map(lambda x: '<p>{0}</p>'.format(escape(x)), input.splitlines()))
    return Markup(''.join(html_p_list))


# In case jinja2 extenions become a need, some good examples can be found at:
#  https://github.com/ckan/ckan/blob/master/ckan/lib/



#!/usr/bin/python
"""
This module implements the application management commands
"""
import shutil
import sys
import os
from glob import glob
from os.path import basename, exists, join, isdir

import quickweb
import quickweb.app
from quickweb.colorhelper import info, warning, success, print_error, print_success
from quickweb.template_manager import download_template
from quickweb import controller, doc


def create(app_directory, template_name, force):
    """
    Create an application using a template
    """
    print("** Creating app on directory %s " % (info(app_directory + "/")))

    if exists(app_directory):
        if force:
            shutil.rmtree(app_directory)
        else:
            print_error(app_directory + ' already exists!')
            print('Use %s if you want to overwrite.' % warning('--force'))
            sys.exit(2)

    download_template(template_name, app_directory)
    doc_path = os.path.dirname(doc.__file__)
    quickweb_required_mask = join(doc_path, '*.md')
    required_files = glob(quickweb_required_mask)
    base_required_files = [basename(x) for x in required_files]
    print("** Adding startup files %s from %s" %
          (info(str(base_required_files)), info(doc_path)))
    for filename in required_files:
        shutil.copy(filename, app_directory)
    print_success('Application successfully created.')
    print_success('You can start it with:')
    print('    ' + success('quickweb run ' + app_directory))
    print_success('Or read about the app structure with:')
    print('    ' + success('more ' + join(app_directory, 'QuickWeb_Application.md')))
    print("**")


def run(app_directory, listener_address, no_logs):
    quickweb.app.run(app_directory, listener_address, no_logs)


def describe(app_directory):
    quickweb.app.run(app_directory, running_describe=True)
    for key, values in controller._app_root.__dict__.items():
        print(key, values)


def setup_cf_deployment(app_directory):
    app_directory = app_directory or os.getcwd()
    app_name = basename(app_directory)
    webroot_dir = join(app_directory, 'webroot')
    if not isdir(webroot_dir):
        print_error("Unable to find webroot directory '%s'" % webroot_dir)
        exit(2)

    manifest_yaml = """\
---
applications:
- name: {0}
  memory: 128M
  buildpack: python_buildpack
  random-route: true # Choose a proper hostname with host: name
""".format(app_name)
    with open(join(app_directory, 'manifest.yaml'), 'w') as manifest_file:
        manifest_file.write(manifest_yaml)
    with open(join(app_directory, 'requirements.txt'), 'w') as requests_file:
        requests_file.write("quickweb>={0}\n".format(quickweb.version()))

    with open(join(app_directory, 'Procfile'), 'w') as procfile:
        procfile.write("web: quickweb run . --no-logs\n")


def setup_docker_deployment(app_directory):
    app_directory = app_directory or os.getcwd()
    app_name = basename(app_directory)
    webroot_dir = join(app_directory, 'webroot')
    if not isdir(webroot_dir):
        print_error("Unable to find webroot directory '%s'" % webroot_dir)
        exit(2)

    dockerfile_txt = """\
FROM python:alpine

WORKDIR /usr/src/app

COPY requirements.txt ./
RUN pip install --no-cache-dir -r requirements.txt

COPY . .

CMD [ "quickweb", "run", "/usr/src/app", "--no-logs" ]
""".format(app_name)
    with open(join(app_directory, 'Dockerfile'), 'w') as manifest_file:
        manifest_file.write(dockerfile_txt)
    with open(join(app_directory, 'requirements.txt'), 'w') as requests_file:
        requests_file.write("quickweb>={0}\n".format(quickweb.version()))



#!/bin/python

import os
import sys
import time
from glob import glob
from os.path import join
import subprocess

TEST_PORT = 8999

os.environ['PORT'] = str(TEST_PORT)
os.environ['TEST_MODE'] = '1'


test_specs = glob(join('test-app', 'tests', '*.yaml'))


subp = subprocess.Popen([sys.executable, '-m', 'quickweb', 'run', 'test-app'], close_fds=True)

print("Waiting 1s for the server warmup")
# Allow 1s for the process to start
time.sleep(1)


for test_filename in test_specs:
    http_unit_test = join('quickweb', 'http_unit_test.py')
    cmd = sys.executable+' %s http://127.0.0.1:%d %s' % (http_unit_test, TEST_PORT, test_filename)
    rc = os.system(cmd)
    if rc != 0:
        subp.terminate()
        exit(2)

subp.terminate()



"""Installs QuickWeb using distutils

Run:
    python setup.py install

to install this package.
"""
from setuptools import setup, find_packages
from os.path import join

name = "QuickWeb"

desc = "Rapid Development Python Web Framework"
long_desc = "QuickWeb is a rapid web development python framework"


classifiers = '''
Development Status :: 1 - Planning
Intended Audience :: Developers
License :: Freely Distributable
License :: OSI Approved :: BSD License
Operating System :: OS Independent
Programming Language :: Python :: 3.7
Topic :: Internet :: WWW/HTTP :: WSGI
Topic :: Internet :: WWW/HTTP :: WSGI :: Application
Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware
Topic :: Internet :: WWW/HTTP :: WSGI :: Server
'''

requirements = open('requirements.txt').read()

setup(
    name=name,
    version=open(join('quickweb', 'version')).readline().strip("\r\n"),
    description=desc,
    long_description=long_desc,
    author='João Pinto',
    author_email='lamego.pinto@gmail.com',
    classifiers=[x for x in classifiers.splitlines() if x],
    install_requires=[x for x in requirements.splitlines() if x],
    url='https://github.com/OpenPipe/QuickWeb',
    packages=find_packages(),
    package_data={'quickweb.doc': ['*']},
    include_package_data=True,
    entry_points={'console_scripts': ['quickweb = quickweb.__main__:main']}
)



#! /usr/bin/env python
"""HTTP Unit Testing

Usage:
    http_unit_test <server_base_url> <test_file>

Options:
    -h --help     Show this screen.
"""
import sys
import yaml
from docopt import docopt
from os.path import dirname, realpath, join
from webtest import TestApp
from colorama import init

init()

# Make sure we use the source directory for imports when running during development
script_dir = join(dirname(realpath(__file__)), '..')
sys.path.insert(0, script_dir)

from quickweb.colorhelper import info, print_success, print_error  # noqa: E402


def validate(response, validation_rules):
    """
    Validates an HTTP response, headers and body (when provided for validation)
    """
    validate_headers = validation_rules.get('headers')
    if validate_headers:
        for header_match in validate_headers:
            (header_name, header_value) = list(header_match.items())[0]
            found_header = False
            for resp_header_name, resp_header_value in response.headerlist:
                if header_name.lower() == resp_header_name.lower():
                    if resp_header_value.lower() != header_value.lower():
                        print()  # previous print does not have end of line
                        print_error("FAILED: %s %s != %s " % (
                            header_name, resp_header_value, header_value)
                            )
                        exit(1)
                    else:
                        found_header = True
                        break
            if not found_header:
                print()  # previous print does not have end of line
                print_error("FAILED: Header %s missing from response" % header_name)
                print_error("Available Headers" + str(response.headerlist))
                exit(1)

    validate_body_list = validation_rules.get('body', [])
    if validate_body_list and isinstance(validate_body_list, str):
        validate_body_list = [validate_body_list]

    last_found_pos = None
    last_validate_body = None
    # webtest resturns the body as bytes() on python3, we need to decode it for find()
    response_body = str(response.body)
    for validate_body in validate_body_list:
        find_pos = response_body.find(validate_body, last_found_pos)
        if find_pos == -1:
            print()  # previous print does not have end of line
            if last_validate_body:
                print_error("FAILED: '%s' was not found after '%s', in body:\n%s " %
                            (validate_body, last_validate_body, response_body)
                            )
            else:
                print_error("FAILED: '%s' is missing from body:\n%s " %
                            (validate_body, response_body)
                            )
            exit(1)
        last_found_pos = find_pos + len(validate_body)
        last_validate_body = validate_body


def run_test(server_base_url, test_filename):
    """ Run test case fro mthe test file """
    print("Runing tests at %s using %s" % (info(server_base_url), info(test_filename)))
    app = TestApp(server_base_url)

    with open(test_filename, 'r') as yaml_file:
        doc = yaml.load_all(yaml_file)
        for test_doc in doc:
            title = test_doc['title']
            padding = (80 - len(title)) * ' '
            print("- %s%s- " % (test_doc['title'], padding), end="")
            test_input = test_doc.get('input')
            if not test_input:
                continue
            # webtest does the status validation based on the status input parameter
            expect_status = test_doc['validate'].get('status')
            response = app.get(test_input['url'], status=expect_status)
            validate(response, test_doc['validate'])
            print_success("OK")


def main():
    """ main function """
    arguments = docopt(__doc__, version='HTTP Unit Testing (latest)')
    run_test(arguments['<server_base_url>'], arguments['<test_file>'])


if __name__ == "__main__":
    main()
